from collections import deque
import json
import os
import sqlite3
import time
import requests
from pathlib import Path

from .settings import Settings

# Stałe – auth_token (CSRF token) oraz ciasteczka
AUTH_TOKEN = ""  # ustaw właściwą wartość
SHM_SESSION = ""  # ustaw właściwą wartość
SHM_USER = ""  # ustaw właściwą wartość

POST_URL = "http://192.168.1.11:8001/auto_tag/import"
BATCH_SIZE = 1000  # maksymalna liczba wierszy w jednej partii


def flatten_graph(graph):
    """
    Dla każdego tagu (klucza) oblicza domknięcie przechodnie –
    czyli zbiór wszystkich tagów osiągalnych w grafie.
    Używamy cache, aby uniknąć wielokrotnych obliczeń.
    """
    cache = {}

    def dfs(tag):
        if tag in cache:
            return cache[tag]
        result = set()
        # Przechodzimy po bezpośrednich konsekwencjach
        for nxt in graph.get(tag, set()):
            result.add(nxt)
            result |= dfs(nxt)
        cache[tag] = result
        return result

    flattened = {}
    for tag in graph.keys():
        flattened[tag] = dfs(tag)
    return flattened


class TagsRepo:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.db_path = os.path.join(
            os.path.dirname(settings.get_settings_path()), "tags.db"
        )
        regenerate = False
        if not Path(self.db_path).is_file():
            regenerate = True
            print(f"Database file not found: {self.db_path}, will regenerate DB")
        self.init_tags_db()
        if regenerate:
            self.regenerate_db()

    def get_conn(self):
        return sqlite3.connect(self.db_path)

    # --- Inicjalizacja bazy tagów ---
    def init_tags_db(self):
        try:
            conn = self.get_conn()
            cursor = conn.cursor()
            tables = {
                "tags": """
                    CREATE TABLE IF NOT EXISTS "tags" (
                        "index" INTEGER,
                        "id" INTEGER,
                        "name" TEXT,
                        "post_count" INTEGER,
                        "category" INTEGER,
                        "created_at" TIMESTAMP,
                        "updated_at" TIMESTAMP,
                        "is_deprecated" INTEGER,
                        "words" TEXT
                    )
                """,
                "tag_aliases": """
                    CREATE TABLE IF NOT EXISTS "tag_aliases" (
                        "index" INTEGER,
                        "alias" TEXT,
                        "tag" TEXT
                    )
                """,
                "tag_closure": """
                    CREATE TABLE IF NOT EXISTS "tag_closure" (
                        antecedent TEXT NOT NULL,
                        consequent TEXT NOT NULL,
                        depth INTEGER NOT NULL,
                        PRIMARY KEY (antecedent, consequent)
                    )
                """,
                "tag_implications": """
                    CREATE TABLE IF NOT EXISTS "tag_implications" (
                        antecedent TEXT NOT NULL,
                        consequent TEXT NOT NULL,
                        PRIMARY KEY (antecedent, consequent)
                    )
                """,
            }
            indexes = {
                "tags": [
                    """CREATE INDEX IF NOT EXISTS ix_tags_index ON tags ("index")""",
                    "CREATE INDEX IF NOT EXISTS tags_index_category ON tags (category)",
                    "CREATE INDEX IF NOT EXISTS tags_index_created_at ON tags (created_at)",
                    "CREATE INDEX IF NOT EXISTS tags_index_id ON tags (id)",
                    "CREATE INDEX IF NOT EXISTS tags_index_is_deprecated ON tags (is_deprecated)",
                    "CREATE INDEX IF NOT EXISTS tags_index_name ON tags (name)",
                    "CREATE INDEX IF NOT EXISTS tags_index_post_count ON tags (post_count)",
                    "CREATE INDEX IF NOT EXISTS tags_index_updated_at ON tags (updated_at)",
                ],
                "tag_aliases": [
                    """CREATE INDEX IF NOT EXISTS ix_tag_aliases_index ON tag_aliases ("index")""",
                    "CREATE INDEX IF NOT EXISTS tag_aliases_index_alias ON tag_aliases (alias)",
                    "CREATE INDEX IF NOT EXISTS tag_aliases_index_tag ON tag_aliases (tag)",
                ],
                "tag_closure": [
                    "CREATE INDEX IF NOT EXISTS idx_closure_antecedent ON tag_closure (antecedent)"
                ],
                "tag_implications": [
                    "CREATE INDEX IF NOT EXISTS idx_implications_antecedent ON tag_implications (antecedent)"
                ],
            }
            for table, create_stmt in tables.items():
                cursor.execute(create_stmt)
            for table, index_list in indexes.items():
                for index_stmt in index_list:
                    cursor.execute(index_stmt)
            conn.commit()
            conn.close()
        except Exception as e:
            print("Błąd przy inicjalizacji bazy tagów:", e)

    def regenerate_db(self):
        # Połączenie z bazą SQLite i pobranie tagów
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute("DELETE FROM tags")
        cursor.execute("DELETE FROM tag_aliases")
        conn.commit()

        rate_limit = 10  # requests per second
        min_interval = 1.0 / rate_limit  # minimum seconds between requests (0.1 sec)

        data_list = []

        page = 0
        while True:
            print(f"Tagi - Pobieranie od id {page}...")
            start_time = time.monotonic()
            url = f"https://danbooru.donmai.us/tags.json?limit=1000&page=a{page}"
            response = requests.get(url)
            if response.status_code != 200:
                print(
                    f"Błąd przy pobieraniu strony {page}: HTTP {response.status_code}"
                )
                break

            data = response.json()
            if not data:
                break
            page = None
            for item in data:
                id = item.get("id")
                if not page:
                    page = id
                name = item.get("name")
                post_count = item.get("post_count")
                category = item.get("category")
                created_at = item.get("created_at")
                updated_at = item.get("updated_at")
                is_deprecated = item.get("is_deprecated")
                words = json.dumps(item.get("words"))
                data_list.append(
                    (
                        id,
                        name,
                        post_count,
                        category,
                        created_at,
                        updated_at,
                        is_deprecated,
                        words,
                    )
                )

            if len(data) < 1000:
                break

            # Calculate elapsed time and sleep if necessary to enforce the rate limit
            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        print(f"Tagi - Pobrano {len(data_list)} tagów...")
        data_list = sorted(data_list, key=lambda x: x[0])
        data_list = [(idx,) + row for idx, row in enumerate(data_list)]

        cursor.executemany(
            """
                           INSERT INTO tags ("index", id, name, post_count, category, created_at, updated_at, is_deprecated, words)
                           VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                           """,
            data_list,
        )
        conn.commit()
        data_list = []

        page = 0
        while True:
            print(f"Aliasy tagów - Pobieranie od id {page}...")
            start_time = time.monotonic()
            url = f"https://danbooru.donmai.us/tag_aliases.json?limit=1000&only=id,antecedent_name,consequent_name&search[status]=active&page=a{page}"
            response = requests.get(url)
            if response.status_code != 200:
                print(
                    f"Błąd przy pobieraniu strony {page}: HTTP {response.status_code}"
                )
                break

            data = response.json()
            if not data:
                break
            page = None
            for item in data:
                id = item.get("id")
                if not page:
                    page = id
                antecedent = item.get("antecedent_name")
                consequent = item.get("consequent_name")
                data_list.append((antecedent, consequent))

            if len(data) < 1000:
                break

            # Calculate elapsed time and sleep if necessary to enforce the rate limit
            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        print(f"Aliasy tagów - Pobrano {len(data_list)} aliasów tagów...")
        data_list = sorted(data_list, key=lambda x: x[0])
        data_list = [(idx,) + row for idx, row in enumerate(data_list)]

        cursor.executemany(
            """
                           INSERT INTO tag_aliases ("index", alias, tag)
                           VALUES (?, ?, ?)
                           """,
            data_list,
        )
        conn.commit()
        data_list = []

        # Pobranie tagów kategorii "character" (category = 4)
        cursor.execute("SELECT name FROM tags WHERE category = 4")
        character_tags = {row[0] for row in cursor.fetchall()}

        # Pobranie tagów kategorii "copyright" (category = 3)
        cursor.execute("SELECT name FROM tags WHERE category = 3")
        copyright_tags = {row[0] for row in cursor.fetchall()}

        # Pobranie tagów kategorii "meta" (category = 5)
        cursor.execute("SELECT name FROM tags WHERE category = 5")
        meta_tags = {row[0] for row in cursor.fetchall()}

        # Pobranie tagów kategorii "artist" (category = 1)
        cursor.execute("SELECT name FROM tags WHERE category = 1")
        artist_tags = {row[0] for row in cursor.fetchall()}

        cursor.execute("DELETE FROM tag_implications")  # Optional: reset table
        cursor.execute("DELETE FROM tag_closure")  # Optional: reset table
        conn.commit()

        # Budujemy strukturę implikacji: słownik, gdzie
        # kluczem jest antecedent_name, a wartością zbiór consequent_name.
        tag_dict = {}

        page = 0
        while True:
            print(f"Implikacje tagów - Pobieranie od id {page}...")
            url = f"https://danbooru.donmai.us/tag_implications.json?limit=1000&page=a{page}"
            response = requests.get(url)
            if response.status_code != 200:
                print(
                    f"Błąd przy pobieraniu strony {page}: HTTP {response.status_code}"
                )
                break

            data = response.json()
            if not data:
                break
            page = None
            for item in data:
                id = item.get("id")
                if not page:
                    page = id
                if item.get("status") != "active":
                    continue

                antecedent = item.get("antecedent_name")
                consequent = item.get("consequent_name")

                # Dodanie prefiksu, jeżeli tag należy do jednej z kategorii
                if antecedent in character_tags:
                    antecedent = f"character:{antecedent}"
                elif antecedent in copyright_tags:
                    antecedent = f"copyright:{antecedent}"
                elif antecedent in meta_tags:
                    antecedent = f"meta:{antecedent}"
                elif antecedent in artist_tags:
                    antecedent = f"artist:{antecedent}"

                if consequent in character_tags:
                    consequent = f"character:{consequent}"
                elif consequent in copyright_tags:
                    consequent = f"copyright:{consequent}"
                elif consequent in meta_tags:
                    consequent = f"meta:{consequent}"
                elif consequent in artist_tags:
                    consequent = f"artist:{consequent}"

                if antecedent not in tag_dict:
                    tag_dict[antecedent] = set()
                tag_dict[antecedent].add(consequent)

            if len(data) < 1000:
                break
            # Calculate elapsed time and sleep if necessary to enforce the rate limit
            elapsed = time.monotonic() - start_time
            if elapsed < min_interval:
                time.sleep(min_interval - elapsed)

        print(f"Implikacje tagów - Pobrano {len(tag_dict)} implikacji tagów...")
        # Batch insert all unique pairs
        for antecedent, consequents in tag_dict.items():
            for consequent in consequents:
                if antecedent != consequent:
                    cursor.execute(
                        "INSERT OR IGNORE INTO tag_implications VALUES (?, ?)",
                        (antecedent, consequent),
                    )
        conn.commit()
        cursor.executemany(
            "INSERT INTO tag_closure VALUES (?, ?, ?)",
            self.build_transitive_closure(tag_dict),
        )
        conn.commit()
        conn.close()

    def build_transitive_closure(self, tag_dict):
        closure = set()
        for antecedent in tag_dict:
            visited = set()
            queue = deque([(antecedent, 0)])

            while queue:
                current_tag, depth = queue.popleft()
                if current_tag in visited:
                    continue
                visited.add(current_tag)

                # Add to closure if not self-reference
                if current_tag != antecedent:
                    closure.add((antecedent, current_tag, depth))

                # Traverse next level
                for next_tag in tag_dict.get(current_tag, []):
                    queue.append((next_tag, depth + 1))

        return closure


# def garbage():
#
#    # Spłaszczenie struktury – obliczenie domknięcia przechodniego
#    flattened_tag_dict = flatten_graph(tag_dict)
#    print("Spłaszczono strukturę implikacji.")
#
#    # Przygotowanie listy wierszy do CSV
#    # Każdy wiersz: (antecedent, consequents jako space-separated string)
#    csv_rows = []
#    for antecedent, consequents in flattened_tag_dict.items():
#        # Sortujemy, żeby wynik był deterministyczny
#        consequents_str = " ".join(sorted(consequents))
#        csv_rows.append((antecedent, consequents_str))
#
#    print(f"Łącznie wierszy do wysłania: {len(csv_rows)}")
#
#    # Konfiguracja ciasteczek do żądania POST
#    cookies = {
#        'shm_session': SHM_SESSION,
#        'shm_user': SHM_USER
#    }
#
#    # Wysyłanie danych w partiach po BATCH_SIZE wierszy
#    for i in range(0, len(csv_rows), BATCH_SIZE):
#        batch = csv_rows[i:i+BATCH_SIZE]
#        # Utworzenie pliku CSV w pamięci
#        output = io.StringIO()
#        writer = csv.writer(output, quoting=csv.QUOTE_ALL)
#        for row in batch:
#            writer.writerow(row)
#        csv_content = output.getvalue()
#        output.close()
#
#        # Przygotowanie danych formularza (z auth_token) oraz pliku CSV
#        data = {
#            'auth_token': AUTH_TOKEN
#        }
#        files = {
#            'auto_tag_file': ('batch.csv', csv_content, 'text/csv')
#        }
#
#        print(f"Wysyłanie batcha wierszy {i+1} - {i+len(batch)}...")
#        post_response = requests.post(POST_URL, data=data, files=files, cookies=cookies, allow_redirects=False)
#        if post_response.status_code in (200, 302):
#            print(f"Batch {i+1}-{i+len(batch)} wysłany pomyślnie.")
#        else:
#            print(f"Błąd przy wysyłaniu batcha {i+1}-{i+len(batch)}: HTTP {post_response.status_code}")
#
#
#    print("Wszystkie dane zostały wysłane.")
